from __future__ import division

kernel_code = """

// Matrix dimensions
// (chosen as multiples of the thread block size for simplicity)
#define WA %(w_a)d // Matrix A width
#define HA %(h_a)d // Matrix A height
#define WB %(w_b)d // Matrix B width
#define HB WA  // Matrix B height
#define WC WB  // Matrix C width
#define HC HA  // Matrix C height

/* Matrix multiplication: C = A * B.
 * Device code.
 */


////////////////////////////////////////////////////////////////////////////////
//! Matrix multiplication on the device: C = A * B
//! WA is A's width and WB is B's width
////////////////////////////////////////////////////////////////////////////////
__kernel void
matrixMul( __global float4* A,
           __global float4* B,
           __global float4* C )
{

    // Declaring and initializing accumulators
    float4 accumulator1 = {0.0f,0.0f,0.0f,0.0f};
    float4 accumulator2 = {0.0f,0.0f,0.0f,0.0f};
    float4 accumulator3 = {0.0f,0.0f,0.0f,0.0f};
    float4 accumulator4 = {0.0f,0.0f,0.0f,0.0f};

    float4 A1, A2, A3, A4;
    float4 B1, B2, B3, B4;


    // Thread index
    //int tx = get_global_id(0);
    //int ty = get_global_id(1);

    //int a_i;
    //int b_j;

    int f4WA = WA/4;
    int f4WB = WB/4;

    int a_i = get_global_id(1);
    int b_j = get_global_id(0);

    int a_start = a_i*f4WA*4;
    int b_start;    

    int k = 0;
    for(; k < f4WA; ++k)
    {
        // Fetching values from A


        A1 = A[a_start+k];  
        A2 = A[a_start+k+f4WA];
        A3 = A[a_start+k+2*f4WA];
        A4 = A[a_start+k+3*f4WA];

        b_start = k*f4WB*4;

        // Fetching values from B
        B1 = B[b_start+b_j];  
        B2 = B[b_start+b_j+f4WB];
        B3 = B[b_start+b_j+2*f4WB];
        B4 = B[b_start+b_j+3*f4WB];

        accumulator1 += A1.xxxx * B1.xyzw + A1.yyyy * B2.xyzw + A1.zzzz * B3.xyzw + A1.wwww * B4.xyzw;
        accumulator2 += A2.xxxx * B1.xyzw + A2.yyyy * B2.xyzw + A2.zzzz * B3.xyzw + A2.wwww * B4.xyzw;
        accumulator3 += A3.xxxx * B1.xyzw + A3.yyyy * B2.xyzw + A3.zzzz * B3.xyzw + A3.wwww * B4.xyzw;
        accumulator4 += A4.xxxx * B1.xyzw + A4.yyyy * B2.xyzw + A4.zzzz * B3.xyzw + A4.wwww * B4.xyzw;
	}

    int index = a_i*f4WB*4+b_j;
    C[index] = accumulator1;
    C[index+f4WB] = accumulator2;
    C[index+2*f4WB] = accumulator3;
    C[index+3*f4WB] = accumulator4;


}

"""

import pyopencl as cl
from time import time
import numpy


ctx = cl.create_some_context()

for dev in ctx.devices:
    assert dev.local_mem_size > 0

#queue = cl.CommandQueue(ctx, 
#        properties=cl.command_queue_properties.PROFILING_ENABLE)  

queue = cl.CommandQueue(ctx)


a_height = 4000
#a_height = 1024
a_width = 200
#a_width = 256
#b_height == a_width

b_height = a_width
b_width = a_height

h_a = numpy.random.rand(a_height, a_width).astype(numpy.float32)
h_b = numpy.random.rand(b_height, b_width).astype(numpy.float32)
h_c = numpy.empty((a_height, b_width)).astype(numpy.float32)


mf = cl.mem_flags
kernel_params = {"w_a":a_width, "h_a":a_height, "w_b":b_width}

kernel = cl.Program(ctx, kernel_code % kernel_params).build().matrixMul

assert a_width % 4 == 0
assert a_height % 4 == 0
assert b_width % 4 == 0
assert b_height == a_width

# kernel(queue, (w, h), tgt, src, numpy.uint32(w), numpy.uint32(h))

# __call__(queue, a_t_buf, a_buf, source.shape)

# args: queue, domain, *args


# Allocate opencl memory
d_a_buf = cl.Buffer(ctx, mf.READ_ONLY, size=h_a.nbytes)
d_b_buf = cl.Buffer(ctx, mf.READ_ONLY, size=h_b.nbytes)
d_c_buf = cl.Buffer(ctx, mf.WRITE_ONLY, size=h_c.nbytes)


# Send matrices to opencl buffers
t1 = time()
write_events = [cl.enqueue_write_buffer(queue, d_b_buf, h_b)]
write_events+= [cl.enqueue_write_buffer(queue, d_a_buf, h_a)]
cl.wait_for_events(write_events)
push_time = time()-t1


t1 = time()
args = [queue, (int(b_width/4),int(a_height/4)), (1,1)]
args+=[d_a_buf]
args+=[d_b_buf]
args+=[d_c_buf]

event = kernel(*args)

event.wait()

t1 = time()

event = kernel(*args)

event.wait()

#print event.profile.end - event.profile.start

gpu_time = time()-t1

t1 = time()

read_events = [cl.enqueue_read_buffer(queue, d_c_buf, h_c)]

# wait on events
cl.wait_for_events(read_events)

pull_time = time()-t1

gpu_total_time = gpu_time+push_time+pull_time

t1 = time()

write_events = [cl.enqueue_write_buffer(queue, d_b_buf, h_b)]
write_events+= [cl.enqueue_write_buffer(queue, d_a_buf, h_a)]

cl.wait_for_events(write_events)

event = kernel(*args)

event.wait()

cl.enqueue_read_buffer(queue, d_c_buf, h_c).wait()

actual_time = time()-t1

ans1=h_c

print "GPU (s) total:", gpu_total_time
print "PUSH ", push_time
print "PULL ", pull_time
print "COMPUTE ", gpu_time
print "ACTUAL ", actual_time

do_cpu = True
	
if do_cpu:
	t1 = time()
	ans2 = numpy.dot(h_a,h_b)

	cpu_time = time()-t1

	print "CPU (s)", cpu_time

	print "GPU speedup: ", cpu_time/gpu_total_time

	print "GPU==CPU:",numpy.allclose(ans1,ans2)
